"""
Copyright 2021 DataRobot, Inc. and its affiliates.

All rights reserved.

DataRobot, Inc. Confidential.

This is unpublished proprietary source code of DataRobot, Inc. and its affiliates.

The copyright notice above does not evidence any actual or intended publication of such source code.

Released under the terms of DataRobot Tool and Utility Agreement.
"""
# flake8: noqa
# Unused imports on purpose

try:
    from datarobot.models.user_blueprints.models import *
    from datarobot import UserBlueprint
    from datarobot import CustomTask
except (ImportError, ModuleNotFoundError):
    print("Ensure DataRobot Python Client >= 2.27")
    raise



"""
Copyright 2021 DataRobot, Inc. and its affiliates.

All rights reserved.

DataRobot, Inc. Confidential.

This is unpublished proprietary source code of DataRobot, Inc. and its affiliates.

The copyright notice above does not evidence any actual or intended publication of such source code.

Released under the terms of DataRobot Tool and Utility Agreement.
"""

from datarobot_bp_workshop.settings import Settings


class FriendlyRepr(object):
    def __friendly_repr__(self):
        return super(FriendlyRepr, self).__repr__()

    def __repr__(self):
        if Settings().is_dev_mode:
            return super(FriendlyRepr, self).__repr__()
        return self.__friendly_repr__()



"""
Copyright 2021 DataRobot, Inc. and its affiliates.

All rights reserved.

DataRobot, Inc. Confidential.

This is unpublished proprietary source code of DataRobot, Inc. and its affiliates.

The copyright notice above does not evidence any actual or intended publication of such source code.

Released under the terms of DataRobot Tool and Utility Agreement.
"""

import copy
import logging

from typing import Dict, Optional, Union
from datarobot import Blueprint

from datarobot_bp_workshop.utils import PrettyString

FAILED_TO_RENDER_WITH_GRAPHVIZ = (
    "Failed to render with `graphviz`. Is `graphviz` installed to the OS?. "
    "Reverting to graphviz string instead."
)

try:
    import graphviz

    GraphvizType = graphviz.Source
except Exception:
    GraphvizType = str

try:
    from IPython.core.display import DisplayHandle

    DisplayGraphviz = DisplayHandle
except Exception:
    DisplayGraphviz = bytes


LOGGER = logging.getLogger("Visualize")


class Visualize(object):
    """ Provides visualization functionality. """

    DATA_NODE = "Data"
    PREDICT_NODE = "Predict"

    border_color = "#c4c8c9"
    node_color = "#20262a"
    predict_color = "#3ca3e8"
    tooltip_color = "#181d20"

    @classmethod
    def to_graphviz_string(
        cls,
        nodes,
        edges,
        graph_name=None,
        node_to_task_lookup=None,
        vertex_context_lookup=None,
        vertical=False,
    ) -> str:
        """ Generate a GraphViz DiGraph string from nodes and edges. """
        nodes = Visualize.format_nodes_to_draw(
            nodes_to_draw=nodes,
            node_to_task_lookup=node_to_task_lookup,
            vertex_context_lookup=vertex_context_lookup,
        )
        digraph_string = Visualize.simple_to_graphviz(
            nodes=nodes, edges=edges, graph_name=graph_name, vertical=vertical,
        )
        return digraph_string

    @classmethod
    def format_nodes_to_draw(
        cls, nodes_to_draw, node_to_task_lookup=None, vertex_context_lookup=None
    ) -> Dict[str, str]:
        """
        Given nodes to use to generate a GraphViz DiGraph string, format them
        such that they are styled as expected to look familiar to the DataRobot UI.
        """
        node_to_task_lookup = node_to_task_lookup or {}
        output = copy.deepcopy(nodes_to_draw)
        for node_to_draw in output:
            if node_to_draw["label"] == "Prediction":
                node_to_draw["label"] = cls.PREDICT_NODE
            node_to_draw["color"] = '"{}"'.format(cls.border_color)
            node_to_draw["fillcolor"] = (
                '"{}"'.format(cls.predict_color)
                if cls.PREDICT_NODE in [node_to_draw["id"], node_to_draw["label"]]
                else '"{}"'.format(cls.node_color)
            )
            node_to_draw["style"] = '"filled,setlinewidth(2)"'
            node_to_draw["shape"] = "box"
            node_to_draw["fontcolor"] = "white"

            cls.handle_validation(
                node_to_draw,
                node_to_task_lookup.get(node_to_draw["id"]),
                vertex_context_lookup,
            )
        return output

    @classmethod
    def handle_validation(
        cls, node_to_draw, task_id=None, vertex_context_lookup=None
    ) -> None:
        """ Add validation to the components to be used to make a graphviz string. """
        if vertex_context_lookup is None:
            return

        if task_id in vertex_context_lookup:
            errors = vertex_context_lookup[task_id].errors or []
            warnings = vertex_context_lookup[task_id].warnings or []

            if len(errors):
                node_to_draw["color"] = "red1"
            elif len(warnings):
                node_to_draw["color"] = "gold1"

            errors_text = ""
            warnings_text = ""
            if len(errors):
                errors_text += (
                    '<tr><td align="left"><font color="red1">Errors:</font>/td></tr>'
                )
                for error in errors:
                    errors_text += '<tr><td align="left"><font color="red1">{}</font></td></tr>'.format(
                        error
                    )

            if len(warnings):
                warnings_text += '<tr><td align="left"><font color="gold1">Warnings:</font></td></tr>'
                for warning in warnings:
                    warnings_text += '<tr><td align="left"><font color="gold1">- {}</font></td></tr>'.format(
                        warning
                    )

            label = node_to_draw["label"]
            node_to_draw["label"] = (
                '<<table border="0" cellborder="0" cellpadding="3" bgcolor="{node_color}">'
                "<tr>"
                '<td align="center" colspan="2">'
                '<font color="white">{label}</font>'
                "</td>"
                "</tr>"
                "{errors}"
                "{warnings}"
                "</table>>"
            ).format(
                node_color=cls.node_color,
                label=label,
                errors=errors_text,
                warnings=warnings_text,
            )

    @classmethod
    def simple_to_graphviz(cls, nodes, edges, graph_name=None, vertical=False) -> str:
        """ Get blueprint chart in graphviz DOT format.

        Returns
        -------
        unicode
            String representation of chart in graphviz DOT language.
        """
        rankdir = "TB" if vertical else "LR"
        digraph = u'digraph "Blueprint Chart" {'
        digraph += u'\n\tgraph [rankdir={}, bgcolor="#0f1315"]'.format(rankdir)
        for node in nodes:
            if not node["label"].startswith('"') and not node["label"].startswith(
                "<<table "
            ):
                node["label"] = '"{}"'.format(node["label"])
            digraph += u'\n\t{id} [{attrs}, fontname="Ubuntu"]'.format(
                id=node["id"],
                attrs=", ".join(
                    "{}={}".format(attr, node[attr])
                    for attr in (set(node.keys()) - {"id"})
                ),
            )
        for edge in edges:
            digraph += u'\n\t{id0} -> {id1} [color="#c4c8c9", weight=5]'.format(
                id0=edge[0], id1=edge[1]
            )
        if graph_name is not None:
            digraph += u'\n\tlabel="{}\n\n"'.format(graph_name)
            digraph += u'\n\tlabelloc="t"'
            digraph += u"\n\tfontcolor = white"
            digraph += u"\n\tfontsize = 18"
            digraph += u'\n\tfontname="Ubuntu"'
        digraph += u"\n}"
        return digraph

    @classmethod
    def show_dr_blueprint(cls, dr_blueprint: Blueprint) -> Optional[DisplayGraphviz]:
        """ Visualize a DataRobot Blueprint object. """
        chart = dr_blueprint.get_chart()
        return Visualize.show(
            Visualize.to_graphviz_string(
                chart.nodes, chart.edges, graph_name=dr_blueprint.model_type
            )
        )

    @classmethod
    def show(
        cls, graphviz_string: str, as_image: bool = True
    ) -> Union[DisplayGraphviz, None, GraphvizType]:
        """
        Visualize a GraphViz string.

        IPython specific drawing functionality.

        Wraps `as_image`.

        Parameters
        ----------
        graphviz_string: string
            The GraphViz string to render
        as_image: bool (Default=True)
            If enabled, will render as a PNG, otherwise will rely on IPython to display
            the `Source` directly. Rendering as a PNG is preferred as it will auto-resize
            to the size of the rendering space provided, while `Source` will overflow.
        """
        if isinstance(graphviz_string, Blueprint):
            return cls.show_dr_blueprint(graphviz_string)

        try:
            to_render = (
                cls.as_image(graphviz_string, catch_exception=False)
                if as_image
                else cls.as_graphviz(graphviz_string, catch_exception=False)
            )
        except Exception:
            LOGGER.error(FAILED_TO_RENDER_WITH_GRAPHVIZ)
            return PrettyString(graphviz_string)

        try:
            from IPython.display import Image, display

            return display(Image(to_render) if as_image else to_render)
        except Exception:
            LOGGER.error(
                "Failed to display. This method requires an IPython environment. "
                "Reverting to `as_image` or `as_graphviz` instead."
            )
            return to_render

    @classmethod
    def as_image(
        cls, graphviz_string: str, catch_exception=True
    ) -> Optional[bytes] or str:
        """
        Used to render the graph as an image.

        Simply wrap in `IPython.display.Image` to display in a notebook,
        or write it to disk.
        """
        try:
            source = cls.as_graphviz(graphviz_string)
            if isinstance(source, graphviz.Source):
                return source.pipe(format="png")
        except Exception:
            if not catch_exception:
                raise
            LOGGER.error(FAILED_TO_RENDER_WITH_GRAPHVIZ)
        return graphviz_string

    @classmethod
    def as_graphviz(cls, graphviz_string: str, catch_exception=True) -> GraphvizType:
        """
        Used to render the graph form of the blueprint using matplotlib.
        """
        try:
            import graphviz

            return graphviz.Source(graphviz_string)
        except Exception:
            if not catch_exception:
                raise
            LOGGER.error(FAILED_TO_RENDER_WITH_GRAPHVIZ)
            return graphviz_string



"""
Copyright 2021 DataRobot, Inc. and its affiliates.

All rights reserved.

DataRobot, Inc. Confidential.

This is unpublished proprietary source code of DataRobot, Inc. and its affiliates.

The copyright notice above does not evidence any actual or intended publication of such source code.

Released under the terms of DataRobot Tool and Utility Agreement.
"""

from typing import Union, List

from .datarobot_client import (
    UserBlueprint,
    GrantAccessControlWithUsernameValidator,
    GrantAccessControlWithIdValidator,
)
from .exceptions import FeatureUnavailable, FailedToShareException
from .utils import RecipientType, Roles, colorize, Colors
from datarobot.errors import ClientError


class SharingInterface(object):
    """ Methods for sharing User Blueprints """

    def share(
        self,
        user_blueprint_id: str,
        usernames: Union[str, List[str]],
        role: str = Roles.CONSUMER,
    ):
        """ Share a User Blueprint with a user or list of other users by username. """
        if not isinstance(usernames, list):
            usernames = [usernames]

        try:
            UserBlueprint.update_shared_roles(
                user_blueprint_id,
                [
                    GrantAccessControlWithUsernameValidator(
                        role=role,
                        share_recipient_type=RecipientType.USER,
                        username=username,
                    )
                    for username in usernames
                ],
            )
            print(colorize("Shared successfuly.", Colors.SUCCESS))
        except ClientError as e:
            if e.status_code != 403:
                data = e.json
                if data and "message" in data and data["message"]:
                    raise FailedToShareException(data["message"])
                raise e
            raise FeatureUnavailable(
                "Unable to share or sharing is not available on this account."
            )

    def share_with_group(
        self,
        user_blueprint_id: str,
        group_ids: Union[str, List[str]],
        role: str = Roles.CONSUMER,
    ):
        """ Share a User Blueprint with a group or list of groups by id. """
        return self._share_with(
            user_blueprint_id=user_blueprint_id,
            recipient_ids=group_ids,
            recipient_type=RecipientType.GROUP,
            role=role,
        )

    def share_with_org(
        self,
        user_blueprint_id: str,
        org_ids: Union[str, List[str]],
        role: str = Roles.CONSUMER,
    ):
        """ Share a User Blueprint with a group or list of organizations by id. """
        self._share_with(
            user_blueprint_id=user_blueprint_id,
            recipient_ids=org_ids,
            recipient_type=RecipientType.ORGANIZATION,
            role=role,
        )

    def _share_with(
        self,
        user_blueprint_id: str,
        recipient_type: str,
        recipient_ids: Union[str, List[str]],
        role: str,
    ):
        if not isinstance(recipient_ids, list):
            recipient_ids = [recipient_ids]
        try:
            UserBlueprint.update_shared_roles(
                user_blueprint_id,
                [
                    GrantAccessControlWithIdValidator(
                        role=role, share_recipient_type=recipient_type, id=recipient_id
                    )
                    for recipient_id in recipient_ids
                ],
            )
            print(colorize("Shared successfuly.", Colors.SUCCESS))
        except ClientError as e:
            if e.status_code != 403:
                data = e.json
                if data and "message" in data and data["message"]:
                    raise FailedToShareException(data["message"])
                raise e
            raise FeatureUnavailable(
                "Unable to share or sharing is not available on this account."
            )



"""
Copyright 2021 DataRobot, Inc. and its affiliates.

All rights reserved.

DataRobot, Inc. Confidential.

This is unpublished proprietary source code of DataRobot, Inc. and its affiliates.

The copyright notice above does not evidence any actual or intended publication of such source code.

Released under the terms of DataRobot Tool and Utility Agreement.
"""
import re
import sys
from typing import Dict, List, Optional, Union

from datarobot import Project, CustomModelVersion
from datarobot.errors import ClientError

from datarobot_bp_workshop.exceptions import CannotInitializeWorkshopException
from .blueprint_graph import BlueprintGraph
from .datarobot_client import (
    UserBlueprint,
    CustomTask,
    UserBlueprintAvailableInput,
    UserBlueprintsInputType,
    UserBlueprintTaskDefinition,
    UserBlueprintTaskLookupEntry,
    ColnameAndType,
    ParamValuePair,
    UserBlueprintCatalogSearch,
)
from .factories import SimpleFactory, TaskFactory
from .settings import Settings
from .task import Task, TaskDict, Feature, FeatureSelection
from .utils import (
    Colors,
    PrettyList,
    TaskOutputMethod,
    colorize,
    HiddenList,
    override_user_agent,
)
from .workshop_interface import WorkshopInterface


class Workshop(WorkshopInterface):
    """The Workshop for building and modifying blueprints."""

    def __init__(
        self,
        auto_initialize=True,
        project_id=None,
        enforce_rate_limit=True,
        user_blueprint_id=None,
    ):
        Settings().check_if_should_upgrade()
        super(Workshop, self).__init__(
            project_id=project_id, enforce_rate_limit=enforce_rate_limit
        )
        self.project = None

        self._task_definitions_by_task_code: Dict[str, UserBlueprintTaskDefinition] = {}
        self._custom_tasks_by_id: Dict[str, UserBlueprintTaskDefinition] = {}
        self._custom_tasks_by_version_id: Dict[str, UserBlueprintTaskDefinition] = {}
        self._custom_task_definitions_by_task_code: Dict[
            str, List[UserBlueprintTaskDefinition]
        ] = {}

        if auto_initialize:
            self.initialize(user_blueprint_id=user_blueprint_id)

    def set_project(self, project_id: Optional[str] = None) -> WorkshopInterface:
        """
        Set the current context of the Workshop to be within a specific project.

        This will provide the ability to select specific columns from a project.
        """
        self._project_id = project_id
        self.initialize()
        return self

    def initialize(
        self, user_blueprint_id: Optional[str] = None, clean: bool = False
    ) -> WorkshopInterface:
        """
        Initialize the tasks in the Workshop. Useful when creating custom tasks mid-session
        or turning on new feature flags.

        Note: Creating a new, or replacing the current, instance of `Workshop`
        will achieve the same.
        """
        override_user_agent(UserBlueprint)

        try:

            if self._project_id:
                self.project = Project(id=self._project_id)

            if self._associated_user_blueprint_id:
                user_blueprint_id = self._associated_user_blueprint_id

            with self.auto_rate_limiter:
                self._available_tasks = UserBlueprint.get_available_tasks(
                    project_id=self._project_id, user_blueprint_id=user_blueprint_id
                )
            self._construct_task_lookups(clean=clean)
            self._task_factory = TaskFactory(self)

            self._dynamic_initialization()
        except Exception:
            error_msg = (
                "Failed to initialize: is the environment properly configured? "
                "See: https://blueprint-workshop.datarobot.com/configuration.html for more info."
            )
            print(error_msg, file=sys.stderr)
            raise CannotInitializeWorkshopException(error_msg)

        return self

    def refresh(self) -> WorkshopInterface:
        """
        Refresh the tasks in the Workshop. Useful when creating custom tasks mid-session
        or turning on new feature flags.

        Note: this is functionally identical to `initialize`.
        """
        return self.initialize()

    def get(
        self,
        user_blueprint_id: str,
        creator: Optional[str] = None,
        last_modifier_name: Optional[str] = None,
    ) -> BlueprintGraph:
        """
        Retrieve a user blueprint using an id.

        Parameters
        ----------
        user_blueprint_id: str
            The id of the User Blueprint
        creator: str, Optional
            The full name of the creator of the user blueprint, if available.
        last_modifier_name: str, Optional
            The full name of the last modifier of the user blueprint, if available.
        """
        with self.auto_rate_limiter:
            return self.blueprint_from_user_blueprint(
                UserBlueprint.get(user_blueprint_id=user_blueprint_id),
                creator=creator,
                last_modifier_name=last_modifier_name,
            )

    def search_blueprints(
        self,
        search: str = None,
        tag: Union[str, List[str]] = None,
        offset: int = 0,
        limit: int = 100,
        show: bool = False,
        as_list: bool = False,
    ) -> List[BlueprintGraph]:
        """ Search for blueprints by keyword """
        if isinstance(tag, list):
            tag = ",".join(tag)

        try:
            search_results = UserBlueprint.search_catalog(
                search=search,
                tag=tag,
                limit=limit,
                offset=offset,
                owner_user_id=None,
                owner_username=None,
                order_by="-created",
            )
            result = self._blueprints_from_search(search_results)

        except ClientError as e:
            if e.status_code != 403:
                raise e
            print(
                "Search is not yet released, falling back to list with client-side search."
            )
            result = self.list(offset=offset, limit=limit, search=search)

        # If the user requests to show the results as a list or visualized...
        consumed = []
        if show or as_list:
            # If we want to show, we need to collect the results too
            for bp in result:
                if show:
                    bp.show()
                consumed.append(bp)

            # Now we can pretty print it
            return PrettyList(consumed)

        return result

    def list(
        self, offset: int = 0, limit: int = 10, search: str = None
    ) -> List[BlueprintGraph]:
        """ Retrieve a list of your personal user blueprints. """
        with self.auto_rate_limiter:
            return self._blueprints_from_search(
                UserBlueprint.list(offset=offset, limit=limit), local_search=search
            )

    def _blueprints_from_search(
        self,
        results: List[Union[UserBlueprint, UserBlueprintCatalogSearch]],
        local_search: str = None,
    ) -> List[BlueprintGraph]:
        """ Retrieve each User Blueprint object and provide a generator for them. """
        for b in results:
            if not local_search or re.match(local_search, b.model_type):
                yield self.get(
                    b.user_blueprint_id,
                    creator=getattr(b, "creator", None),
                    last_modifier_name=getattr(b, "last_modifier_name", None),
                )

    def clone(
        self,
        blueprint_id: str,
        project_id: Optional[str] = None,
        name: Optional[str] = None,
    ) -> BlueprintGraph:
        """ Clone a blueprint from a project and save it as a user blueprint. """
        if project_id is None:
            project_id = self._project_id

        if project_id is None:
            raise ValueError(
                "Project Id must be provided if not set for the `Workshop`."
            )

        with self.auto_rate_limiter:
            return self.blueprint_from_user_blueprint(
                UserBlueprint.clone_project_blueprint(
                    blueprint_id=blueprint_id, project_id=project_id, model_type=name
                )
            )

    def delete(self, *user_blueprint_ids: str):
        """ Delete user blueprints using ids. """
        for user_blueprint_id in user_blueprint_ids:
            with self.auto_rate_limiter:
                UserBlueprint.delete(user_blueprint_id=user_blueprint_id)
        print(colorize("Blueprints deleted.", Colors.SUCCESS))

    def list_categories(
        self,
        categories: List[str] = None,
        show_tasks: bool = False,
        color: bool = True,
        margin: str = "",
    ) -> HiddenList:
        """ List the categories of tasks, and optionally the tasks themselves. """
        categories = (
            self._available_tasks.categories if categories is None else categories
        )
        seen_custom_task_id = set()
        for category in categories:

            name = category.name
            if color:
                name = colorize(name, Colors.BLUE)
            print("{}{}\n".format(margin, name))
            # This is a tree structure: show any subcategories first
            self.list_categories(
                color=color,
                margin="{}  ".format(margin),
                categories=category.subcategories or [],
                show_tasks=show_tasks,
            )

            if show_tasks:
                items = []
                for task_code in category.task_codes:
                    task = self._get_task(task_code, fallback_to_none=True)
                    # For example, a custom task...
                    if task is not None:
                        items += ["  - {} ({})".format(task.label, task_code)]
                    elif task_code in self._custom_task_definitions_by_task_code:
                        for definition in self._custom_task_definitions_by_task_code[
                            task_code
                        ]:
                            if definition.custom_task_id in seen_custom_task_id:
                                continue
                            seen_custom_task_id.add(definition.custom_task_id)
                            task = self._get_custom_task(
                                definition.custom_task_id, fallback_to_none=True
                            )
                            item = "  - {} ({})".format(
                                task.label,
                                "{}_{}".format(task_code, definition.custom_task_id),
                            )
                            items += [item]

                print("{}".format(margin) + "\n{}".format(margin).join(items))

        return HiddenList(categories)

    def search_tasks(
        self, term: str, ignore_case: bool = True, fuzzy: bool = True
    ) -> PrettyList:
        """
        Search for tasks which match a given term, optionally
        ignoring case (default) or performing a fuzzy search (default).
        """
        flags = 0
        if ignore_case:
            flags = re.IGNORECASE

        def match(task):
            code = (
                task.task_code
                if hasattr(task, "custom_task_id")
                else task.custom_task_id
            )
            return re.match(
                ".*".join([""] + term.split(" ") + [""] if fuzzy else term.split(" ")),
                " ".join(
                    [task.label, task.description, code, " ".join(task.categories)]
                ),
                flags,
            )

        built_in_tasks = [
            self._get_task(task.task_code, fallback_to_none=True)
            for task in self._task_definitions_by_task_code.values()
            if match(task)
        ]

        custom_tasks = [
            self._get_custom_task(task.custom_task_id)
            for task in self._custom_tasks_by_id.values()
            if match(task)
        ]

        tasks = sorted(built_in_tasks + custom_tasks, key=lambda t: str(t))
        return PrettyList([t for t in tasks if t is not None])

    def BlueprintGraph(
        self, final_task: Task, name: Optional[str] = None
    ) -> BlueprintGraph:
        """
        Provides a mechanism to declare a set of tasks in the form of a directed acyclic graph.

        Parameters
        ----------
        final_task: Task
            The task to use as the final task in a blueprint, which will be inspected
            to find all inputs.
        name: str
            Specify a name to use for the blueprint
        """
        return BlueprintGraph(workshop=self, final_task_or_tasks=final_task, name=name)

    def Task(
        self,
        task_code: str,
        output_method: Optional[str] = None,
        task_parameters: Union[List[ParamValuePair], dict] = None,
        output_method_parameters: Union[List[ParamValuePair], dict] = None,
        x_transformations: Union[List[ParamValuePair], dict] = None,
        y_transformations: Union[List[ParamValuePair], dict] = None,
        freeze: bool = False,
    ) -> Task:
        """
        Provides a mechanism to specify how to transform or provide predictions for input
        data.
        """
        task = Task(
            workshop=self,
            task_code=task_code,
            output_method=output_method,
            task_parameters=task_parameters,
            output_method_parameters=output_method_parameters,
            x_transformations=x_transformations,
            y_transformations=y_transformations,
            freeze=freeze,
        )
        return task

    def CustomTask(
        self,
        custom_task_id: str,
        output_method=None,
        task_parameters: List[ParamValuePair] = None,
        output_method_parameters: List[ParamValuePair] = None,
        x_transformations: List[ParamValuePair] = None,
        y_transformations: List[ParamValuePair] = None,
        freeze: bool = False,
        version: Optional[str] = None,
    ) -> Task:
        """ Create a custom task, using a `custom_task_id`, and optionally a version. """
        # Let the user pass CUSTOMR_<id> or <id>
        custom_task_id = self._parse_custom_id(custom_task_id)
        version = self._parse_custom_id(version)

        task = Task(
            workshop=self,
            task_code=self._custom_tasks_by_id[custom_task_id].task_code,
            output_method=output_method,
            task_parameters=task_parameters,
            output_method_parameters=output_method_parameters,
            x_transformations=x_transformations,
            y_transformations=y_transformations,
            freeze=freeze,
            custom_task_id=custom_task_id,
            version=version,
        )
        return task

    def Feature(self, feature_name: str) -> Feature:
        """
        Selects a feature from the project data.

        Ensure to call `set_project` or reinitialize the `Workshop` with a `project_id`,
        before attempting to select a feature.

        Parameters
        ----------
        feature_name: str
            The feature to select from the project data.
        """
        return Feature(self, feature_name)

    def FeatureSelection(
        self, *feature_names: Union[str, ColnameAndType], exclude=False
    ) -> FeatureSelection:
        """
        Selects a number of features from the project data of a specific data type.

        Ensure to call `set_project` or reinitialize the `Workshop` with a `project_id`,
        before attempting to select a feature.

        Parameters
        ----------
        feature_names: *Union[str, ColnameAndType]
            Include features which match the list provided.
        exclude: bool
            Whether to exclude the list of provided features, instead of include.
        """
        return FeatureSelection(self, feature_names, exclude=exclude)

    def deserialize_blueprint(self, bp_json: List[Dict]) -> BlueprintGraph:
        """ Create a BlueprintGraph from a list of dictionaries representing a BlueprintGraph. """
        return BlueprintGraph.deserialize(workshop=self, bp_json=bp_json)

    def deserialize_task(
        self, task_json: TaskDict, original_id: Optional[str] = None
    ) -> Task:
        """ Create a Task from a dictionary representing Task. """
        return Task.deserialize(
            workshop=self, task_dict=task_json, original_id=original_id,
        )

    def blueprint_from_user_blueprint(
        self,
        user_blueprint: UserBlueprint,
        creator: Optional[str] = None,
        last_modifier_name: Optional[str] = None,
    ) -> BlueprintGraph:
        """
        Create a BlueprintGraph from a UserBlueprint object

        Parameters
        ----------
        user_blueprint: UserBlueprint
        creator: str, Optional
            The full name of the creator of the user blueprint, if available.
        last_modifier_name: str, Optional
            The full name of the last modifier of the user blueprint, if available.

        Returns
        -------
        BlueprintGraph
        """
        return BlueprintGraph.from_user_blueprint(
            self, user_blueprint, creator=creator, last_modifier_name=last_modifier_name
        )

    def add_to_repository(
        self, project_id: str, user_blueprint_id: str
    ) -> Optional[str]:
        """Add a user blueprint to a project's repository

        Parameters
        ----------
        project_id: str
            the id of the project
        user_blueprint_id: str
            the id of the user blueprint

        Returns
        -------
        str or None
            blueprint_id if the user blueprint was successfully
            added to the repository. None otherwise.
        """
        with self.auto_rate_limiter:
            response = UserBlueprint.add_to_project(
                project_id=project_id, user_blueprint_ids=[user_blueprint_id]
            )
        added = response.added_to_menu
        if not added:
            return UserBlueprint.get(user_blueprint_id).blueprint_id
        return added[0].blueprint_id

    def link_to_project(self, user_blueprint_id: str, project_id: Optional[str]):
        """
        Allows linking or unlinking of a user blueprint from a project.
        Especially useful for unlinking a blueprint from a project you no longer have access to.

        When a user blueprint is linked to a project, it will be able to validate and otherwise use
        project specific tasks, such as feature selection.

        If a user blueprint is linked to a project, it is expected that it will only be used with
        that project, or clones of that project.

        If a user blueprint is not linked to a project, it is expected that it should be able to be
        project-agnostic, that is, used with any project which shares the same target type as the
        user blueprint.

        Parameters
        ----------
        user_blueprint_id: str
            Id of the User Blueprint
        project_id: str or None
            Id of the project to link to (or None).
            If None is passed, unlink the user blueprint from the project
        """
        user_blueprint = UserBlueprint.update(
            None,
            user_blueprint_id=user_blueprint_id,
            project_id=project_id,
            include_project_id_if_none=True,
        )
        return BlueprintGraph.from_user_blueprint(self, user_blueprint)

    def create_custom_task(
        self, environment_id: str, name: str, target_type: str, description: str = None
    ):
        """
        Create a custom task, which can then be updated using code via %%update_custom
        or via the DataRobot API.

        Parameters
        ----------
        environment_id: str
            ID of the environment to use to create the model
        name: str
            The name of the custom task
        target_type: str
            One of: ("Anomaly", "Binary", "Multiclass", "Regression", "Transform")
        description: Optional[str]
            An optional description of the model

        Returns
        -------
        CustomTask
        """
        custom_task = CustomTask.create(
            name=name, target_type=target_type, description=description
        )
        _ = CustomModelVersion.create_from_previous(
            custom_task.id, environment_id, is_major_update=False
        )
        self.initialize()
        return custom_task

    def get_custom_task(self, custom_task_id: str):
        """ Retrieve a custom task. """
        return CustomTask.get(custom_task_id)

    def _dynamic_initialization(self):
        """ Perform dynamic initialization of objects, classes, and methods. """
        stock_tasks: List[UserBlueprintTaskLookupEntry] = []
        custom_tasks: List[UserBlueprintTaskLookupEntry] = []
        columns_and_types: List[ColnameAndType] = []

        for task in self._available_tasks.tasks:
            task_definition: UserBlueprintTaskDefinition = task.task_definition
            if task_definition.custom_task_id is None:
                stock_tasks += [task]
            else:
                custom_tasks += [task]

            if not columns_and_types and task_definition.colnames_and_types:
                columns_and_types = task_definition.colnames_and_types

        self._colname_and_type_colname_lookup = {
            t.colname: t for t in columns_and_types
        }
        self._colname_and_type_hex_lookup = {t.hex: t for t in columns_and_types}

        # Tasks
        self.Tasks = self._build_tasks(stock_tasks)
        self.TaskCodes = self._build_tasks_codes(stock_tasks)

        # Custom Tasks
        self.CustomTasks = self._build_custom_tasks(custom_tasks)

        # Inputs
        with self.auto_rate_limiter:
            input_types: UserBlueprintAvailableInput = UserBlueprint.get_input_types()
        self._input_name_lookup = {i.type: i.name for i in input_types.input_types}
        self.TaskInputs = self._build_input_types(input_types.input_types)
        self.TaskOutputMethod = TaskOutputMethod

        # Feature Lookup
        self.Features = self._build_feature_data(columns_and_types)
        self.feature_lookup = {item.colname: item.hex for item in columns_and_types}

    def _parse_custom_id(self, custom_id: Optional[str]):
        if not custom_id or not isinstance(custom_id, str):
            return custom_id

        if "_" in custom_id:
            for possible in custom_id.split("_"):
                if possible in self._custom_tasks_by_id:
                    custom_id = possible
        return custom_id

    def _construct_task_lookups(self, clean=False):
        if clean:
            self._task_definitions_by_task_code = {}
            self._custom_tasks_by_id = {}
            self._custom_tasks_by_version_id = {}
            self._custom_task_definitions_by_task_code = {}

        for task in self._available_tasks.tasks:
            self._task_argument_lookup[task.task_code] = {
                a.key: a.argument for a in task.task_definition.arguments
            }
            custom_task_id = task.task_definition.custom_task_id
            if custom_task_id:
                self._custom_task_definitions_by_task_code.setdefault(
                    task.task_code, []
                ).append(task.task_definition)
                self._custom_tasks_by_id[custom_task_id] = task.task_definition
                for version in task.task_definition.custom_task_versions:
                    self._custom_tasks_by_version_id[version.id] = task.task_definition
            else:
                self._task_definitions_by_task_code[
                    task.task_code
                ] = task.task_definition
        return self._available_tasks

    def _get_task(self, task_code: str, fallback_to_none: bool = False) -> Task:
        if fallback_to_none and not hasattr(self.Tasks, task_code):
            return None
        return getattr(self.Tasks, task_code)

    def _get_custom_task_key(self, task: UserBlueprintTaskLookupEntry) -> str:
        return self._custom_task_key_string(task.task_definition.custom_task_id)

    def _get_custom_task(
        self, custom_task_id: str, fallback_to_none: bool = False
    ) -> Task:
        custom_task_string = self._custom_task_key_string(custom_task_id)
        if fallback_to_none and not hasattr(self.CustomTasks, custom_task_string):
            return None
        return getattr(self.CustomTasks, custom_task_string)

    def _build_tasks_codes(self, tasks: List[UserBlueprintTaskLookupEntry]):
        return SimpleFactory(tasks)(
            get_key=lambda i: i.task_code,
            get_value=lambda i: i.task_code,
            get_type=lambda i: "string",
            get_description=lambda i: i.task_definition.description,
            title="Unique Codes of the Available Tasks in the Workshop",
            class_name="TaskCodes",
        )

    def _build_tasks(self, tasks: List[UserBlueprintTaskLookupEntry]):
        return SimpleFactory(tasks)(
            get_key=lambda i: i.task_code,
            get_value=lambda i: self._task_factory(i.task_definition),
            get_type=lambda i: "Callable -> Task",
            get_description=lambda i: i.task_definition.description,
            title="Available Tasks in the Workshop",
            class_name="Tasks",
        )

    def _build_custom_tasks(self, tasks: List[UserBlueprintTaskLookupEntry]):
        return SimpleFactory(tasks)(
            get_key=self._get_custom_task_key,
            get_value=lambda i: self._task_factory(i.task_definition),
            get_type=lambda i: "Callable -> Task",
            get_description=lambda i: i.task_definition.description,
            title="Available Custom Tasks in the Workshop",
            class_name="CustomTasks",
        )

    def _build_input_types(self, input_types: List[UserBlueprintsInputType]):
        return SimpleFactory(input_types)(
            get_key=lambda i: i.type,
            get_value=lambda i: i.type,
            get_type=lambda i: "string",
            get_description=lambda i: i.name,
            title="Available Input Types in the Workshop",
            class_name="TaskInputs",
        )

    def _build_feature_data(self, columns_and_types: List[ColnameAndType]):
        return SimpleFactory(columns_and_types)(
            get_key=lambda i: i.colname,
            get_value=lambda i: self.Feature(i.colname),
            get_type=lambda i: "Feature",
            get_description=lambda i: "{} input data from the feature: '{}'".format(
                self._input_name_lookup[i.type], i.colname
            ),
            title="Specific Features Available in the Associated Project",
            class_name="Features",
        )



"""
Copyright 2021 DataRobot, Inc. and its affiliates.

All rights reserved.

DataRobot, Inc. Confidential.

This is unpublished proprietary source code of DataRobot, Inc. and its affiliates.

The copyright notice above does not evidence any actual or intended publication of such source code.

Released under the terms of DataRobot Tool and Utility Agreement.
"""

from typing import List

from .datarobot_client import (
    UserBlueprintTaskArgument,
    UserBlueprintTaskCustomTaskMetadata,
    UserBlueprint,
)


class UserBlueprintHelper(object):
    @staticmethod
    def get_custom_task_version_ids(user_blueprint: UserBlueprint):
        version_ids = set()
        for task in user_blueprint.blueprint:
            for task_param in task.task_data.task_parameters:
                if task_param.param == "version_id":
                    version_ids.add(task_param.value)
        return version_ids

    @staticmethod
    def get_task_codes(user_blueprint: UserBlueprint):
        """ Retrieve all task codes, not including custom tasks """
        return {
            task.task_data.task_code
            for task in user_blueprint.blueprint
            if "version_id"
            not in {task_param.param for task_param in task.task_data.task_parameters}
        }

    @staticmethod
    def get_arguments(arguments: List[UserBlueprintTaskArgument]):
        return [
            a
            for a in sorted(arguments, key=lambda a: a.argument.name)
            if a.argument.tunable is not False and not a.key.startswith("pp_")
        ]

    @staticmethod
    def get_docstring_for_parameter(
        argument: UserBlueprintTaskArgument, minified=False
    ):
        key = argument.key
        value = argument.argument
        template = "{name} ({key}): {type}, (Default={default})\n"
        if not minified:
            template += "    Possible Values: {values}\n"
        return template.format(
            name=value.name,
            key=key,
            type=value.type,
            default=value.default.__repr__(),
            values=value.values,
        )

    @staticmethod
    def get_docstring_for_arguments(
        arguments: List[UserBlueprintTaskArgument], title: str, minified=False
    ):
        return UserBlueprintHelper.get_docstring_for_lines(
            lines=[
                UserBlueprintHelper.get_docstring_for_parameter(
                    entry, minified=minified
                )
                for entry in arguments
            ],
            title=title,
            minified=minified,
        )

    @staticmethod
    def get_docstring_for_versions(
        versions: List[UserBlueprintTaskCustomTaskMetadata], title: str, minified=False
    ):
        template = "{} ({}): str\n"
        if not minified:
            template += (
                "    The version id used to determine "
                "which version of a custom task to use.\n"
            )
        return UserBlueprintHelper.get_docstring_for_lines(
            lines=[template.format(version.label, version.id) for version in versions],
            title=title,
            minified=minified,
        )

    @staticmethod
    def get_docstring_for_lines(lines: List[str], title: str, minified=False):
        docstring_definitions = []

        if not minified:
            if title:
                docstring_definitions = [title, ""]

            docstring_definitions += ["Parameters", "----------"]

        docstring_definitions += lines

        return "\n".join(docstring_definitions)



"""
Copyright 2021 DataRobot, Inc. and its affiliates.

All rights reserved.

DataRobot, Inc. Confidential.

This is unpublished proprietary source code of DataRobot, Inc. and its affiliates.

The copyright notice above does not evidence any actual or intended publication of such source code.

Released under the terms of DataRobot Tool and Utility Agreement.
"""

import json
import os
from typing import Optional
from urllib import request

from ._version import __version__

# Global Session Information
SESSION_INFO = {"VERSION_CHECK_ATTEMPTED": False}


class Settings(object):
    """ Introduce concept of settings """

    force_no_color = False
    force_dev_mode = False
    force_no_version_check = False

    defaults = {
        "BLUEPRINT_WORKSHOP_DEV_MODE": False,
        "BLUEPRINT_WORKSHOP_ALLOW_COLOR": True,
        "BLUEPRINT_WORKSHOP_VERSION_CHECK": True,
    }

    @property
    def is_dev_mode(self) -> bool:
        return self.force_dev_mode or self.get_env_value("BLUEPRINT_WORKSHOP_DEV_MODE")

    @property
    def allow_color(self) -> bool:
        if self.force_no_color:
            return False
        return self.get_env_value("BLUEPRINT_WORKSHOP_ALLOW_COLOR")

    @property
    def should_check_version(self) -> bool:
        if SESSION_INFO["VERSION_CHECK_ATTEMPTED"]:
            return False
        if self.force_no_version_check:
            return False
        return self.get_env_value("BLUEPRINT_WORKSHOP_VERSION_CHECK")

    @classmethod
    def get_env_value(cls, setting):
        return os.environ.get(setting, cls.defaults[setting])

    @staticmethod
    def get_latest_version() -> Optional[str]:
        """ Parse the version info JSON from PyPI """
        version_info = {}
        try:
            with request.urlopen(
                "https://pypi.org/pypi/datarobot-bp-workshop/json"
            ) as f:
                version_info = json.loads(f.read().decode("utf-8"))
        except Exception:
            # Silently fail if unable to perform version check
            # This failure is likely due to no network connection.
            pass

        return (
            version_info.get("info", {}).get("version")
            if version_info is not None
            else None
        )

    def is_latest_version_or_none(self) -> bool:
        """ Easily testable version comparison check """
        version = self.get_latest_version()
        # If version cannot be obtained, resolve as if latest
        return True if version is None else version == __version__

    def check_if_should_upgrade(self) -> bool:
        """ Check PyPI for the latest version and recommend update. """
        should_upgrade = False
        if self.should_check_version:
            # Check to see if latest version is being used
            if not self.is_latest_version_or_none():
                should_upgrade = True
                print(
                    "Please upgrade to the latest version: "
                    "pip install --upgrade datarobot_bp_workshop"
                )
            # We attempted to check - wait until next session to try again
            self.set_version_check_attempted(True)
        return should_upgrade

    @staticmethod
    def set_version_check_attempted(attempted: bool):
        """
        Set whether the version check has been attempted

        Parameters
        ----------
        attempted: bool
            Whether the version check has been attempted
        """
        SESSION_INFO["VERSION_CHECK_ATTEMPTED"] = attempted



"""
Copyright 2021 DataRobot, Inc. and its affiliates.

All rights reserved.

DataRobot, Inc. Confidential.

This is unpublished proprietary source code of DataRobot, Inc. and its affiliates.

The copyright notice above does not evidence any actual or intended publication of such source code.

Released under the terms of DataRobot Tool and Utility Agreement.
"""


class CycleException(Exception):
    """ Error raised if a cycle is discovered. """


class TaskFrozenError(Exception):
    """ Raised when a frozen task is requested to be modified. """


class CustomTaskNotFound(Exception):
    """ Raised when a custom task id is provided but is unavailable. """


class NoProjectSpecifiedException(Exception):
    """ Raised if there is no project specified to be used for training. """


class UnsavedBlueprintException(Exception):
    """ Blueprint must be saved to perform ths operation. """


class FeatureUnavailable(Exception):
    """ To be raised when a feature is not available on the current account. """


class CannotInitializeWorkshopException(Exception):
    """ Raised when blueprint-workshop has a configuration error and cannot use the DataRobot API. """


class FailedToShareException(Exception):
    """ Raised when a user blueprint is unable to be shared. """



"""DataRobot Blueprint Workshop
================================

The Blueprint Workshop is built to ensure that constructing and modifying DataRobot blueprints
through a programmatic interface is idiomatic, convenient, and powerful.

This library requires a `DataRobot`_ account.

The easiest way to get started is by reading the
`Documentation`_.

This package is covered under the DataRobot Tool and Utility Agreement, which can be
read on our `Legal`_ page, along with our privacy policy and more.

Prerequisites
-----------------
Python >= 3.7

Ensure your DataRobot account has enabled:

- Show Uncensored Blueprints
- Enable Customizable Blueprints

Installation
-----------------
Optionally, create a virtual environment

::

    $ mkvirtualenv -p python3.7 blueprint-workshop


Install Graphviz on Linux (to render blueprints):

::

    $ sudo apt-get install graphviz


Install Graphviz on Mac (to render blueprints):

::

    $ brew install graphviz


Install the library!

::

    $ pip install datarobot-bp-workshop


Recommended:

::

    $ pip install jupyterlab


From a folder where you'd like to save your scripts:

::

    $ jupyter-lab .

.. _datarobot: https://datarobot.com
.. _documentation: https://blueprint-workshop.datarobot.com
.. _legal: https://www.datarobot.com/legal/
"""

import re
import sys

from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand


class PyTest(TestCommand):
    user_options = [("pytest-args=", "a", "Arguments to pass to pytest")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = ""

    def run_tests(self):
        import shlex

        # import here, cause outside the eggs aren't loaded
        import pytest

        errno = pytest.main(shlex.split(self.pytest_args))
        sys.exit(errno)


with open("datarobot_bp_workshop/_version.py") as fd:
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', fd.read(), re.MULTILINE
    ).group(1)

if not version:
    raise RuntimeError("Cannot find version information")

lint_require = [
    "flake8==3.9",
    'black==19.10b0; python_version >= "3.6"',
]

tests_require = [
    "pytest-mock==3.5.1",
    "pytest==6.1.2",
    "pytest-check==1.0.1",
    "pytest-cov==2.12.1",
]

docs_require = [
    "Sphinx==3.5.4",
    "sphinx_rtd_theme==0.5.2",
    "numpydoc>=0.6.0",
    "nbsphinx>=0.8.5,<1",
    "sphinx-tabs==3.1.0",
    "sphinx-rtd-dark-mode==1.2.2",
    "ipython==7.25.0",
]

dev_require = tests_require + lint_require + docs_require + ["twine==1.6.5"]

release_require = ["zest.releaser[recommended]==6.22.0"]

setup(
    name="datarobot-bp-workshop",
    version=version,
    description="This client library is designed to support building DataRobot blueprints.",
    author="datarobot",
    author_email="support@datarobot.com",
    maintainer="datarobot",
    maintainer_email="info@datarobot.com",
    url="https://blueprint-workshop.datarobot.com",
    license="DataRobot Tool and Utility Agreement",
    packages=find_packages("."),
    long_description=__doc__,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "License :: Other/Proprietary License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    install_requires=["datarobot==2.27.0", "networkx>=2.2", "graphviz==0.16"],
    tests_require=tests_require,
    extras_require={
        "dev": dev_require,
        "docs": docs_require,
        "release": release_require,
        "lint": lint_require,
    },
    cmdclass={"test": PyTest},
)
